﻿using System;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;

using Pong.Behaviours;
using Apos.Shapes;

using Syntriax.Engine.Core;
using Syntriax.Engine.Physics2D;
using System.Linq;
using Syntriax.Engine.Network;
using Syntriax.Engine.Core.Factory;

namespace Pong;

[System.Serializable]
public class TestMessagePacket : INetworkPacket, LiteNetLib.Utils.INetSerializable
{
    public string Message { get; set; }

    public void Deserialize(LiteNetLib.Utils.NetDataReader reader) => Message = reader.GetString();
    public void Serialize(LiteNetLib.Utils.NetDataWriter writer) => writer.Put(Message);

    public TestMessagePacket() => Message = "Default Message";
    public TestMessagePacket(string Message) => this.Message = Message;
}

public class GamePong : Game
{
    private readonly IUniverseObject platformSpecificUniverseObject = null!;
    private readonly GraphicsDeviceManager graphics = null!;
    private SpriteBatch spriteBatch = null!;
    private ShapeBatch shapeBatch = null!;

    private Universe universe = null!;
    private BehaviourCollector<IDisplayableSprite> displayableCollector = null!;
    private BehaviourCollector<IDisplayableShape> displayableShapeCollector = null!;
    private MonoGameCamera2DBehaviour cameraBehaviour = null!;

    private PongManagerBehaviour pongManager = null!;
    private TextScoreBehaviour leftText;
    private TextScoreBehaviour rightText;

    public GamePong(IUniverseObject platformSpecificUniverseObject)
    {
        this.platformSpecificUniverseObject = platformSpecificUniverseObject;
        graphics = new GraphicsDeviceManager(this)
        {
            PreferredBackBufferWidth = 1024,
            PreferredBackBufferHeight = 576,
            GraphicsProfile = GraphicsProfile.HiDef
        };

        Content.RootDirectory = "Content";
        IsMouseVisible = true;
    }

    protected override void Initialize()
    {
        // TODO: Add your initialization logic here
        universe = new();
        displayableCollector = new(universe);
        displayableShapeCollector = new(universe);

        universe.Initialize();

        base.Initialize();
    }

    protected override void LoadContent()
    {
        spriteBatch = new SpriteBatch(GraphicsDevice);
        shapeBatch = new ShapeBatch(GraphicsDevice, Content);
        SpriteFont spriteFont = Content.Load<SpriteFont>("UbuntuMono");

        universe.Register(platformSpecificUniverseObject);

        NetworkManager networkManager = universe.InstantiateUniverseObject<NetworkManager>().SetUniverseObject("Network Manager");
        networkManager.NetworkCommunicator = universe.InstantiateUniverseObject<LiteNetLibClient>().SetUniverseObject("Client").Connect("localhost", 8888).SendToServer(new TestMessagePacket("Hola")).SubscribeToPackets<TestMessagePacket>(OnClientPacketReceived);
        if (Environment.GetCommandLineArgs().FirstOrDefault(x => x.CompareTo("-server") == 0) is not null)
        {
            universe.InstantiateUniverseObject<LiteNetLibServer>().SetUniverseObject("Server").Start(8888, 2).SubscribeToPackets<TestMessagePacket>(OnServerPacketReceived);
            Window.Title = $"{Window.Title} - Server";
        }
        universe.InstantiateUniverseObject<PhysicsEngine2D>().SetUniverseObject("Physics Engine 2D");

        ////////////////////////////////////////////////////////////////////////////////////

        cameraBehaviour = universe.InstantiateUniverseObject().SetUniverseObject("Camera")
            .BehaviourController.AddBehaviour<Transform2D>()
            .BehaviourController.AddBehaviour<CameraController>()
            .BehaviourController.AddBehaviour<MonoGameCamera2DBehaviour>(graphics);

        ////////////////////////////////////////////////////////////////////////////////////

        pongManager = universe.InstantiateUniverseObject().SetUniverseObject("Pong Game Manager")
            .BehaviourController.AddBehaviour<PongManagerBehaviour>(5);

        ////////////////////////////////////////////////////////////////////////////////////

        universe.InstantiateUniverseObject().SetUniverseObject("Ball")
            .BehaviourController.AddBehaviour<Transform2D>().SetTransform(position: new Vector2D(0, 0f), scale: new Vector2D(10f, 10f))
            .BehaviourController.AddBehaviour<CircleBehaviour>(new Circle(Vector2D.Zero, 1f))
            .BehaviourController.AddBehaviour<BallBehaviour>()
            .BehaviourController.AddBehaviour<RigidBody2D>();

        ////////////////////////////////////////////////////////////////////////////////////

        IUniverseObject leftPaddle = UniverseObjectFactory.Instantiate().SetUniverseObject("Left Paddle");
        leftPaddle.Id = "lp";
        leftPaddle.BehaviourController.AddBehaviour<Transform2D>().SetTransform(position: new Vector2D(-468f, 0f), scale: new Vector2D(15f, 60f))
            .BehaviourController.AddBehaviour<PaddleBehaviour>(Keys.W, Keys.S, 228f, -228f, 400f)
            .BehaviourController.AddBehaviour<ShapeBehaviour>(Shape2D.Square)
            .BehaviourController.AddBehaviour<RigidBody2D>().IsStatic = true;
        universe.Register(leftPaddle);

        IUniverseObject rightPaddle = UniverseObjectFactory.Instantiate().SetUniverseObject("Right Paddle");
        rightPaddle.Id = "rp";
        rightPaddle.BehaviourController.AddBehaviour<Transform2D>().SetTransform(position: new Vector2D(468f, 0f), scale: new Vector2D(15f, 60f))
            .BehaviourController.AddBehaviour<PaddleBehaviour>(Keys.Up, Keys.Down, 228f, -228f, 400f)
            .BehaviourController.AddBehaviour<ShapeBehaviour>(Shape2D.Square)
            .BehaviourController.AddBehaviour<RigidBody2D>().IsStatic = true;
        universe.Register(rightPaddle);

        ////////////////////////////////////////////////////////////////////////////////////

        universe.InstantiateUniverseObject().SetUniverseObject("Wall Top")
            .BehaviourController.AddBehaviour<Transform2D>().SetTransform(position: new Vector2D(0f, 308f), scale: new Vector2D(552f, 20f))
            .BehaviourController.AddBehaviour<ShapeBehaviour>(Shape2D.Square)
            .BehaviourController.AddBehaviour<RigidBody2D>().IsStatic = true;

        universe.InstantiateUniverseObject().SetUniverseObject("Wall Bottom")
            .BehaviourController.AddBehaviour<Transform2D>().SetTransform(position: new Vector2D(0f, -308f), scale: new Vector2D(552f, 20f))
            .BehaviourController.AddBehaviour<ShapeBehaviour>(Shape2D.Square)
            .BehaviourController.AddBehaviour<RigidBody2D>().IsStatic = true;

        universe.InstantiateUniverseObject().SetUniverseObject("Wall Right")
            .BehaviourController.AddBehaviour<Transform2D>().SetTransform(position: new Vector2D(532f, 0f), scale: new Vector2D(20f, 328f))
            .BehaviourController.AddBehaviour<WallScoreBehaviour>((Action)pongManager.ScoreToLeft)
            .BehaviourController.AddBehaviour<ShapeBehaviour>(Shape2D.Square)
            .BehaviourController.AddBehaviour<RigidBody2D>().IsStatic = true;

        universe.InstantiateUniverseObject().SetUniverseObject("Wall Left")
            .BehaviourController.AddBehaviour<Transform2D>().SetTransform(position: new Vector2D(-532f, 0f), scale: new Vector2D(20f, 328f))
            .BehaviourController.AddBehaviour<WallScoreBehaviour>((Action)pongManager.ScoreToRight)
            .BehaviourController.AddBehaviour<ShapeBehaviour>(Shape2D.Square)
            .BehaviourController.AddBehaviour<RigidBody2D>().IsStatic = true;

        ////////////////////////////////////////////////////////////////////////////////////

        leftText = universe.InstantiateUniverseObject().SetUniverseObject("Score Left")
            .BehaviourController.AddBehaviour<Transform2D>().SetTransform(position: new Vector2D(-250f, 250f), scale: Vector2D.One * .25f)
            .BehaviourController.AddBehaviour<TextScoreBehaviour>(true, spriteFont);

        rightText = universe.InstantiateUniverseObject().SetUniverseObject("Score Right")
            .BehaviourController.AddBehaviour<Transform2D>().SetTransform(position: new Vector2D(250f, 250f), scale: Vector2D.One * .25f)
            .BehaviourController.AddBehaviour<TextScoreBehaviour>(false, spriteFont);
    }

    private void OnClientPacketReceived(TestMessagePacket packet, string from)
    {
        Console.WriteLine($"Message received from server: {packet.Message}");
        leftText.Text = packet.Message;
    }

    private void OnServerPacketReceived(TestMessagePacket packet, string from)
    {
        Console.WriteLine($"Message received from client: {packet.Message}");
        rightText.Text = packet.Message;

        universe.GetRequiredUniverseObject<INetworkCommunicatorServer>().SendToClient(from, new TestMessagePacket("Reply!"));
    }

    protected override void Update(GameTime gameTime)
    {
        if (GamePad.GetState(PlayerIndex.One).Buttons.Back == ButtonState.Pressed || Keyboard.GetState().IsKeyDown(Keys.Escape))
            Exit();

        universe.Update(gameTime.ToUniverseTime());

        if (Keyboard.GetState().IsKeyDown(Keys.S))
            universe.GetRequiredUniverseObject<INetworkCommunicatorClient>().SendToServer<TestMessagePacket>(new TestMessagePacket($"Hola ({gameTime.TotalGameTime.TotalSeconds})"));

        base.Update(gameTime);
    }

    protected override void Draw(GameTime gameTime)
    {
        GraphicsDevice.Clear(new Color() { R = 32, G = 32, B = 32 });

        // TODO: Add your drawing code here
        universe.PreDraw();

        spriteBatch.Begin(SpriteSortMode.Deferred, transformMatrix: cameraBehaviour.MatrixTransform);
        foreach (var displayable in displayableCollector)
            displayable.Draw(spriteBatch);
        spriteBatch.End();

        shapeBatch.Begin(cameraBehaviour.MatrixTransform);
        foreach (var displayableShape in displayableShapeCollector)
            displayableShape.Draw(shapeBatch);
        shapeBatch.End();

        base.Draw(gameTime);
    }
}
