using System;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using Syntriax.Engine.Core;

namespace Syntriax.Engine.Network;

public class NetworkManager : UniverseObject, INetworkManager
{
    private INetworkCommunicator networkCommunicator = null!;
    private readonly List<(Type packetType, Delegate callback)> delegates = [];

    public INetworkCommunicator NetworkCommunicator
    {
        get => networkCommunicator;
        set
        {
            if (networkCommunicator == value)
                return;

            var previousCommunicator = networkCommunicator;
            networkCommunicator = value;

            if (previousCommunicator is not null) UnsubscribeDelegates(networkCommunicator);
            if (networkCommunicator is not null) SubscribeDelegates(networkCommunicator);
        }
    }

    private readonly Dictionary<string, INetworkEntity> _networkEntities = [];
    public IReadOnlyDictionary<string, INetworkEntity> NetworkEntities => _networkEntities;

    private readonly BehaviourCollector<INetworkEntity> _networkEntityCollector = new();
    public IBehaviourCollector<INetworkEntity> NetworkEntityCollector => _networkEntityCollector;

    public NetworkManager()
    {
        CacheDelegates();

        _networkEntityCollector.OnCollected += OnCollected;
        _networkEntityCollector.OnRemoved += OnRemoved;
    }

    private void CacheDelegates()
    {
        // Find network packets implementing EntityDataPacket<>
        IEnumerable<Type> packetTypes = AppDomain.CurrentDomain.GetAssemblies().SelectMany(a => a.GetTypes())
            .Where(t => typeof(IEntityNetworkPacket).IsAssignableFrom(t) && !t.IsInterface && !t.IsAbstract && !t.IsGenericType);

        MethodInfo onPacketArrivedMethod = GetType()
            .GetMethod(nameof(OnPacketReceived), BindingFlags.NonPublic | BindingFlags.Instance)!;

        foreach (Type packetType in packetTypes)
        {
            MethodInfo genericOnPacketArrivedMethod = onPacketArrivedMethod.MakeGenericMethod(packetType);

            Type genericDelegateType = typeof(Action<,>).MakeGenericType(packetType, typeof(string));
            Delegate genericPacketReceivedDelegate = Delegate.CreateDelegate(genericDelegateType, this, genericOnPacketArrivedMethod);

            delegates.Add((packetType, genericPacketReceivedDelegate));
        }
    }

    private void SubscribeDelegates(INetworkCommunicator networkCommunicator)
    {
        MethodInfo subscribeToPacketsMethod = typeof(INetworkCommunicator)
            .GetMethod(nameof(INetworkCommunicator.SubscribeToPackets), BindingFlags.NonPublic | BindingFlags.Public | BindingFlags.Instance)!;

        foreach ((Type packetType, Delegate callback) in delegates)
        {
            MethodInfo genericSubscribeMethod = subscribeToPacketsMethod.MakeGenericMethod(packetType);

            genericSubscribeMethod.Invoke(networkCommunicator, [callback]);
        }
    }

    private void UnsubscribeDelegates(INetworkCommunicator networkCommunicator)
    {
        MethodInfo unsubscribeFromPacketsMethod = typeof(INetworkCommunicator)
            .GetMethod(nameof(INetworkCommunicator.UnsubscribeFromPackets), BindingFlags.NonPublic | BindingFlags.Public | BindingFlags.Instance)!;

        foreach ((Type packetType, Delegate callback) in delegates)
        {
            MethodInfo genericUnsubscribeMethod = unsubscribeFromPacketsMethod.MakeGenericMethod(packetType);

            genericUnsubscribeMethod.Invoke(networkCommunicator, [callback]);
        }
    }

    private void OnPacketReceived<T>(T entityDataPacket, string fromClientId)
    {
        if (entityDataPacket is IEntityNetworkPacket entityNetworkPacket)
            _networkEntities[entityNetworkPacket.EntityId].ReceiveDataClient(entityDataPacket);
    }

    private void OnCollected(IBehaviourCollector<INetworkEntity> sender, INetworkEntity behaviourCollected)
    {
        if (!_networkEntities.TryAdd(behaviourCollected.Id, behaviourCollected))
            throw new($"Unable to add {behaviourCollected.Id} to {nameof(NetworkManager)}");
    }

    private void OnRemoved(IBehaviourCollector<INetworkEntity> sender, INetworkEntity behaviourRemoved)
    {
        _networkEntities.Remove(behaviourRemoved.Id);
    }

    protected override void OnEnteringUniverse(IUniverse universe) => _networkEntityCollector.Assign(universe);
    protected override void OnExitingUniverse(IUniverse universe) => _networkEntityCollector.Unassign();
}
